(ns com.aurellem.assembly
  (:use (com.aurellem gb-driver vbm title items))
  (:import [com.aurellem.gb_driver SaveState]))

(defn mid-game []
  (read-state "mid-game"))

(defn inject-assembly
  ([^SaveState state
   program-counter registers
   assembly-code]
  (let [scratch-memory (memory state)]
    ;; inject assembly code
    (dorun (map (fn [index val]
                  (aset scratch-memory index val))
                (range program-counter
                       (+ program-counter (count assembly-code)))
                assembly-code))
    (-> state
        (write-memory! scratch-memory)
        (write-registers! registers)
        (PC! program-counter)))))

(defn inject-item-assembly
  ([^SaveState state assembly-code]
     (inject-assembly state (inc item-list-start)
                      (registers state)
                      assembly-code))
  ([assembly-code]
     (inject-item-assembly @current-state assembly-code)))

(defn info
  ([^SaveState state]
     (println (format "PC: 0x%04X" (PC state)))
     (println "Instruction:"
              (format "0x%02X" (aget (memory state) (PC state))))
     state))    

(defn print-interrupt
  [^SaveState state]
  (println (format "IE: %d" (IE state)))
  state)

(defn run-assembly
  ([info-fn assembly n]
     (let [final-state
           (reduce (fn [state _]
                     (tick (info-fn state)))
                   (inject-item-assembly
                     (mid-game) assembly)
                   (range n))]
       final-state))
  ([assembly n]
     (run-assembly info assembly n)))

(def buttons-port 0xFF00)

(defn A [state]
  (bit-shift-right (bit-and 0x0000FF00 (AF state)) 8))

(defn binary-str [num]
  (format "%08d"
          (Integer/parseInt
           (Integer/toBinaryString num) 10)))

(defn view-register [state name reg-fn]
  (println (format "%s: %s" name
                   (binary-str (reg-fn state))))
  state)


(defn view-memory [state mem]
  (println (format "mem 0x%04X = %s" mem
                   (binary-str (aget (memory state) mem))))
  state)

(defn read-down-button []
  (-> (tick (mid-game))
      (IE! 0) ; disable interrupts
      (inject-item-assembly
       (concat
        ;; write 00010000 to 0xFF00 to select joypad
        [0x18   ;D31D                    ; jump over          
         0x01   ;D31E                    ; the next 8 bits
                ;D31F
         (Integer/parseInt "00100000" 2) ; data section 
         
         0xFA   ;D320                    ; load (D31F) into A
         0x1F   ;D321      -->       
         0xD3   ;D322      -->  D31F     

         0xEA   ;D323                    ; load (A), which is 
         0x00   ;D324      -->           ; 00010000, into FF00
         0xFF   ;D325      -->  FF00     
         
         0x18   ;D326                    ; this is the place where
         0x01   ;D327                    ; we will store whether
         0x00   ;D328                    ; "down" is pressed.

         0xFA   ;D329                    ; (FF00) -> A
         0x00   ;D32A                   
         0xFF   ;D32B

         0xCB   ;D32C                    ; Test whether "down"
         0x5F   ;D32D                    ; is pressed.

         0x28   ;D32E                    ; if down is pressed,
         0x03   ;D32F                    ; skip the next section 
                                         ; of code.
         ;; down-is-not-pressed
         0xC3   ;D330
         0x1D   ;D331                    ; return to beginning
         0xD3   ;D332
         
         ;; down-is-pressed 
         0xEA   ;D334                    ; write A to D328 if 
         0x28   ;D335                    ; "down" was pressed
         0xD3   ;D336

         0xC3   ;D330
         0x1D   ;D331                    ; return to beginning
         0xD3   ;D332
         ]
        
        []))))


;; specs for main bootstrap program
;; starts in "mode-select" mode
;;   Each button press takes place in a single frame.
;;   mode-select-mode takes one of the main buttons
;;   which selects one of up to eight modes
;;   mode 1 activated by the "A" button
;;   the next two button presses indicates the start
;;   memory location which to which the bootstrap
;;   program will write.
;;   This is done by using each of the eight buttons to
;;   spell out an 8 bit number.  The order of buttons is
;;   ["A" "B" "start" "select" "up" "right" "down" "left"],
;;   [:a :start :l]  -->  10100001

;;   the next button press determines how many bytes are to be
;;   written, starting at the start position.

;;   then, the actual bytes are entered and are written to the
;;   start address in sequence.


(defn count-frames []
  (-> (tick (mid-game))
      (IE! 0) ; disable interrupts
      (inject-item-assembly
       ;; write 00010000 to 0xFF00 to select joypad
       [0x18   ;D31D                    ; jump over          
        0x02   ;D31E                    ; the next 2 bytes
        0x00   ;D31F                    ; frame-count
        0x00   ;D320                    ; v-blank-prev
        
        0xFA   ;D321
        0x41   ;D322                    ; load (FF41) into A
        0xFF   ;D323                    ; this contains mode flags


        
        ;; if we're in v-blank, the bit-1 is 0
        ;; and bit-2 is 1  Otherwise, it is not v-blank.
        0xCB   ;D324                     ; test bit-1 of A
        0x4F   ;D325                         

        0xC2   ;D326                     ; if bit-1 is not 0
        0x43   ;D327                     ; GOTO not-v-blank
        0xD3   ;D328
        
        0xCB   ;D329                     ; test bit-0 of A 
        0x47   ;D32A

        0xCA   ;D32B                     ; if bit-0 is not 1
        0x43   ;D32C                     ; GOTO not-v-blank
        0xD3   ;D32D
        
        ;;; in v-blank mode

           ;; if v-blank-prev was 0,
           ;; increment frame-count

        0xFA   ;D32E                    ; load v-blank-prev to A
        0x20   ;D32F
        0xD3   ;D330
        
        0xCB   ;D331
        0x47   ;D332                    ; test bit-0 of A 

        0x20   ;D333                    ; skip next section
        0x07   ;D334                    ; if v-blank-prev was not zero 
        
           ;; v-blank was 0, increment frame-count
        0xFA   ;D335                    ; load frame-count into A
        0x1F   ;D336
        0xD3   ;D337                   

        0x3C   ;D338                    ; inc A

        0xEA   ;D339                    ; load A into frame-count
        0x1F   ;D33A
        0xD3   ;D33B

           ;; set v-blank-prev to 1
        0x3E   ;D33C                    ; load 1 into A
        0x01   ;D33D                    

        0xEA   ;D33E                    ; load A into v-blank-prev
        0x20   ;D33F
        0xD3   ;D340

        0x18   ;D341                    ; skip not-in-v-blank section
        0x05   ;D342

        ;;; not in v-blank mode
           ;; set v-blank-prev to 0
        0x3E   ;D343                    ; load 0 into A        
        0x00   ;D344

        0xEA   ;D345                    ; load A into v-blank-prev
        0x20   ;D346
        0xD3   ;D347
        

        0xC3   ;D348                   ; return to beginning
        0x1D   ;D349
        0xD3   ;D34A
        ])))

(defn step-count-frames []
  (-> (read-down-button)
      (info)
      (tick)  ;; skip over data section
      (info)  
      (view-register "Register A" A)
      (tick)  ;; load-data into A
      (view-register "Register A" A)
      (info)
      (view-memory 0xFF00)
      (tick) ;; load A into 0xFF00
      (view-memory 0xFF00)
      (info)
      (tick)
      (info)
      (tick)
      (info)
      (tick)
      (info)
      (tick)
      (info)
      (tick)
      (info)
      (tick)
      (print-inventory)))

(defn test-read-down []
  (= (view-memory (step (step (read-buttons) [:d])) 0xD328)
     (view-memory (step (step (read-buttons))) 0xD328)))
  
(defn trace [state]
  (loop [program-counters []
         opcodes []]
    (let [frame-boundary?
          (com.aurellem.gb.Gb/tick)]
      (println (count opcodes))
      (if frame-boundary?
        [program-counters opcodes]
        (recur
         (conj program-counters
               (first (registers @current-state)))
         (conj opcodes
               (aget (memory @current-state)
                     (PC @current-state))))))))
  
(defn good-trace []
  (-> (mid-game) (tick) (IE! 0)
      (set-inv-mem [0x00 0x00 0X00 0x00])
      (PC! item-list-start)(print-interrupt)
      (info) (tick) (info) (tick) (info)))









